"""
Geographical measurement and simple analysis module for Python. Provides
Point, Multipoint, Line, and Polygon classes, with methods for simple
measurements such as distance, area, and bearing.
"""
from __future__ import division

import math
import sys
import traceback
import numpy as np
from . import vtk
from . import geojson
from . import xyfile
from . import _shpfuncs
from .. import crs as kcrs
from .metadata import Metadata
from . import _vectorgeo

try:
    from . import _cvectorgeo as _vecgeo
except ImportError:
    sys.stderr.write("falling back on slow _vectorgeo")
    _vecgeo = _vectorgeo

try:
    import pyproj
    PYPROJ = True
    geod = pyproj.Geod(ellps="WGS84")
except ImportError:
    PYPROJ = False

class Geometry(object):
    """ This is the abstract base class for all geometry types """
    _geotype = None

    def __init__(self, crs=kcrs.CARTESIAN):
        self.properties = {}
        self._crs = crs
        return

    def _distance(self, pos0, pos1):
        """ Generic method for calculating distance between positions that
        respects CRS """
        if self._crs == kcrs.CARTESIAN:
            dist = _vecgeo.distance((pos0.x, pos0.y), (pos1.x, pos1.y))
        elif PYPROJ:
            _, _, dist = geod.inv(pos0.x, pos0.y, pos1.x, pos1.y, radians=False)
        else:
            dist = greatcircle(pos0, pos1)
        return dist

    def add_property(self, name, value):
        """ Insert a property (name -> value) into the properties dict, raising
        a NameError if the property already exists. Compared to directly
        manipulating the properties dict, this avoids accidents. """
        if name not in self.properties:
            self.properties[name] = value
        else:
            raise NameError("property '{0}' already exists in {1}".format(name, type(self)))
        return

    def delete_property(self, name):
        """ Equivalent to

            del self.properties[name]
        """
        if name in self.properties:
            del self.properties[name]
        return


class Point(Geometry):
    """ Point object instantiated with:

    *coords*        2-tuple or 3-tuple
    *data*          List, dictionary, or Metadata object for point-specific
                    data [default None]
    *properties*    Dictionary of geometry specific data [default None]
    *crs*           Coordinate reference system instance [default CARTESIAN]
    """
    _geotype = "Point"

    def __init__(self, coords, data=None, properties=None, copy_metadata=True,
                 **kwargs):
        if not hasattr(coords, "__iter__"):
            raise ValueError("Point coordinates must be a sequence")
        super(Point, self).__init__(**kwargs)
        self.vertex = coords
        self._setxyz()
        self.data = Metadata(data, singleton=True, copydata=copy_metadata)
        if hasattr(properties, "keys"):
            self.properties = properties
        else:
            properties = {}
        return

    def __getitem__(self, idx):
        return self.vertex[idx]

    def __repr__(self):
        return 'Point(' + str(self.vertex) + ')'

    def __eq__(self, other):
        try:
            return (tuple(self.vertex) == tuple(other.vertex)) and \
                   (self.data == other.data) and \
                   (self.properties == other.properties)
        except AttributeError:
            return False

    @property
    def __geo_interface__(self):
        return {"type" : "Point", "coordinates" : self.vertex}

    def _setxyz(self):
        """ Create convenience *x*, *y* (, *z*) attributes from vertices. """
        vertex = self.vertex
        self.x = vertex[0]
        self.y = vertex[1]
        if len(vertex) == 2:
            self.rank = 2
            self.z = None
        else:
            self.rank = 3
            self.z = vertex[2]
        return

    def isat(self, other, tol=0.001):
        """ Test whether point vertices are the same, regardless of properties or data.
        Tolarance is not geodetic. """
        return False not in [abs(a-b) <= tol for (a, b) in zip(self.vertices,
                                                               other.vertices)]

    def get_vertex(self):
        """ Return the Point vertex as a tuple. """
        return self.vertex

    def coordsxy(self, convert_to=False):
        """ Returns the x,y coordinates. Convert_to may be set to 'deg'
        or 'rad' for convenience.  """
        if convert_to == 'rad':
            return (self.x*3.14159/180., self.y*3.14159/180.)
        elif convert_to == 'deg':
            return (self.x/3.14159*180., self.y/3.14159*180.)
        else:
            return (self.x, self.y)

    def azimuth(self, other, spherical=False):
        """ Returns the compass azimuth from self to other in radians (i.e.
        clockwise, with north at 0). Returns NaN if points are coincident. """

        if self.coordsxy() == other.coordsxy():
            return np.nan

        if self._crs == kcrs.LONLAT and other._crs == kcrs.LONLAT:
            if not PYPROJ:
                raise CRSError("Non-cartesian points require pyproj")
            az1, _, _ = geod.inv(self.x, self.y, other.x, other.y)
            return az1 * math.pi / 180.0

        elif self._crs == kcrs.CARTESIAN and other._crs == kcrs.CARTESIAN:
            dx = other.x - self.x
            dy = other.y - self.y

            if dx > 0:
                if dy > 0:
                    return math.atan(dx/dy)
                elif dy < 0:
                    return math.pi - math.atan(-dx/dy)
                else:
                    return 0.5*math.pi
            elif dx < 0:
                if dy > 0:
                    return 2*math.pi - math.atan(-dx/dy)
                elif dy < 0:
                    return math.pi + math.atan(dx/dy)
                else:
                    return 1.5*math.pi
            else:
                if dy > 0:
                    return 0.0
                else:
                    return math.pi

        else:
            raise CRSError("Azimuth undefined for points in CRS {0} and "
                           "{1}".format(self._crs, other._crs))

    def walk(self, distance, direction):
        """ Returns the point reached when moving in a given direction for
        a given distance from a specified starting location.

            distance (float): distance to walk
            direction (float): horizontal walk direction in radians
        """
        if self._crs == kcrs.CARTESIAN:
            dx = distance * math.sin(direction)
            dy = distance * math.cos(direction)

            if self.rank == 2:
                return Point((self.x+dx, self.y+dy),
                             properties=self.properties, data=self.data, crs=self._crs)
            elif self.rank == 3:
                return Point((self.x+dx, self.y+dy, self.z),
                             properties=self.properties, data=self.data, crs=self._crs)

        elif PYPROJ:
            (x, y, backaz) = geod.fwd(self.x, self.y, direction*180.0/math.pi,
                                      distance, radians=False)
            return Point((x, y), properties=self.properties, data=self.data,
                         crs=self._crs)

        else:
            raise CRSError("Non-cartesian points require pyproj")

    def distance(self, other):
        """ Returns a distance to another Point. If the coordinate system is
        geographical and a third (z) coordinate exists, it is assumed to have
        the same units as the real-world horizontal distance (i.e. meters). """
        if self._crs != other._crs:
            raise CRSError("Points must share the same coordinate system.")
        flat_dist = self._distance(self, other)
        if None in (self.z, other.z):
            return flat_dist
        else:
            return math.sqrt(flat_dist**2. + (self.z-other.z)**2.)

    def greatcircle(self, other):
        """ Return the great circle distance between two geographical points. """
        if not PYPROJ:
            raise CRSError("Non-cartesian points require pyproj")
        if not (self._crs == kcrs.LONLAT and other._crs == kcrs.LONLAT):
            raise CRSError("Great circle distances require both points to be "
                           "in geographical coordinates")
        az1, az2, dist = geod.inv(self.x, self.y, other.x, other.y, radians=False)
        return dist

    def shift(self, shift_vector):
        """ Shift point by the amount given by a vector. Operation occurs
        in-place """
        if len(shift_vector) != self.rank:
            raise GGeoError('Shift vector length must equal geometry rank.')

        self.vertex = tuple([a+b for a,b in zip(self.vertex, shift_vector)])
        self._setxyz()
        return self

    def as_geojson(self, **kwargs):
        """ Write data as a GeoJSON string to a file-like object `f`.

        Parameters
        ----------
        f : file-like object to recieve the GeoJSON string

        *kwargs* include:
        crs : coordinate reference system
        crs_fmt : format of `crs`; may be one of ('epsg','ogc_crs_urn')
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        writer = geojson.GeoJSONWriter(self, **kwargs)
        return writer.print_json()

    def to_geojson(self, f, **kwargs):
        """ Write data as a GeoJSON string to a file-like object `f`.

        Parameters
        ----------
        f : file-like object to recieve the GeoJSON string

        *kwargs* include:
        crs : coordinate reference system
        crs_fmt : format of `crs`; may be one of ('epsg','ogc_crs_urn')
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        try:
            if not hasattr(f, "write"):
                fobj = open(f, "w")
            else:
                fobj = f
            writer = geojson.GeoJSONWriter(self, **kwargs)
            writer.write_json(fobj)
        finally:
            if not hasattr(f, "write"):
                fobj.close()
        return writer


class MultipointBase(Geometry):
    """ Point cloud with associated attributes. This is a base class for the
    polyline and polygon classes. """
    _geotype = "MultipointBase"

    def __init__(self, vertices, data=None, properties=None, copy_metadata=True,
                 **kwargs):
        super(MultipointBase, self).__init__(**kwargs)
        vertices = list(vertices)
        if len(vertices) > 0:
            self.rank = len(vertices[0])

            if not 2 <= self.rank <= 3:
                raise GInitError("Input must be doubles or triples")
            elif not all(self.rank == len(i) for i in vertices):
                raise GInitError("Input must have consistent rank")
            else:
                self.vertices = [tuple(i) for i in vertices]

        else:
            self.rank = None
            self.vertices = []

        if hasattr(properties, "keys"):
            self.properties = properties
        elif properties is None:
            self.properties = {}
        else:
            raise GInitError("value provided as 'properties' must be a hash")

        self.data = Metadata(data, copydata=copy_metadata)
        return

    def __repr__(self):
        if len(self) < 5:
            ppverts = str(self.vertices)
        else:
            ppverts = str(self.vertices[:2])[:-1] + "..." + str(self.vertices[-2:])[1:]
        return '{typ}({verts})>'.format(
                typ=str(type(self))[:-1], verts=ppverts)

    def __len__(self):
        return len(self.vertices)

    def __getitem__(self, key):
        if isinstance(key, (int, np.int64)):
            return Point(self.vertices[key], data=self.data[key],
                         properties=self.properties, crs=self._crs,
                         copy_metadata=False)
        elif isinstance(key, slice):
            return type(self)(self.vertices[key], data=self.data[key],
                              properties=self.properties, crs=self._crs,
                              copy_metadata=False)
        else:
            raise GGeoError('Index must be an integer or a slice object')

    def __setitem__(self, key, value):
        if not isinstance(key, int):
            raise GGeoError('Indices must be integers')
        if getattr(value, "_geotype", None) == "Point":
            self.vertices[key] = value.vertex
            self.data[key] = list(value.data.values())[0]
        elif len(value) == self.rank:
            self.vertices[key] = value
            self.data[key] = None
        else:
            raise GGeoError('Cannot insert non-Pointlike value with '
                            'length != {0}'.format(self.rank))
        return

    def __delitem__(self, key):
        if len(self) > key:
            del self.vertices[key]
            del self.data[key]
        else:
            raise GGeoError('Index ({0}) exceeds length'
                            '({1})'.format(key, len(self)))
        return

    def __iter__(self):
        return (self[i] for i in range(len(self)))

    def __eq__(self, other):
        return hasattr(other, "_geotype") and \
               (self._geotype == other._geotype) and \
               (len(self) == len(other)) and \
               all(a==b for a,b in zip(self, other))

    def _bbox_overlap(self, other):
        """ Return whether bounding boxes between self and another geometry
        overlap.
        """
        reg0 = self.bbox
        reg1 = other.bbox
        return (reg0[0] < reg1[2] and reg0[2] > reg1[0] and
                reg0[1] < reg1[3] and reg0[3] > reg1[1])

    @property
    def bbox(self):
        """ Return the extents of a bounding box as
            (xmin, ymin, xmax, ymax)
        """
        if self.rank == 2:
            x, y = self.get_coordinate_lists()
        elif self.rank == 3:
            x, y, z = self.get_coordinate_lists()
        bbox = (min(x), min(y), max(x), max(y))
        return bbox

    def print_vertices(self):
        """ Prints an enumerated list of indices. """
        for i, vertex in enumerate(self.vertices):
            print("{0}\t{1}".format(i, vertex))

    def get_vertices(self):
        """ Return vertices as a list of tuples. """
        return np.array(self.vertices)

    def get_coordinate_lists(self):
        """ Return X, Y, and Z lists. If self.rank == 2, Z will be
        zero-filled. """
        X = [i[0] for i in self.vertices]
        Y = [i[1] for i in self.vertices]
        if self.rank == 3:
            Z = [i[2] for i in self.vertices]
            return X, Y, Z
        else:
            return X, Y

    def shift(self, shift_vector):
        """ Shift feature by the amount given by a vector. Operation
        occurs in-place """
        if len(shift_vector) != self.rank:
            raise GGeoError('Shift vector length must equal geometry rank.')

        if self.rank == 2:
            f = lambda pt: (pt[0] + shift_vector[0], pt[1] + shift_vector[1])
        elif self.rank == 3:
            f = lambda pt: (pt[0] + shift_vector[0], pt[1] + shift_vector[1],
                            pt[2] + shift_vector[2])
        self.vertices = list(map(f, self.vertices))
        return self

    def _matmult(self, A, x):
        """ Return Ax=b """
        b = []
        for a in A:
            b.append(sum([ai * xi for ai, xi in zip(a, x)]))
        return b

    def rotate2d(self, thetad, origin=(0, 0)):
        """ Rotate rank 2 Multipoint around *origin* counter-clockwise by
        *thetad* degrees. """
        # First, shift by the origin
        self.shift([-a for a in origin])

        # Multiply by a rotation matrix
        theta = thetad / 180.0 * math.pi
        R = ((math.cos(theta), -math.sin(theta)),
             (math.sin(theta), math.cos(theta)))
        rvertices = [self._matmult(R, x) for x in self.vertices]
        self.vertices = rvertices

        # Shift back
        self.shift(origin)
        return self

    def _subset(self, idxs):
        """ Return a subset defined by index in *idxs*. """
        vertices = [self.vertices[i] for i in idxs]
        data = self.data.sub(idxs)
        subset = type(self)(vertices, data=data, properties=self.properties,
                            crs=self._crs, copy_metadata=False)
        return subset

    def flat_distances_to(self, pt):
        """ Return the "flat Earth" distance from each vertex to a point. """
        A = np.array(self.vertices)
        P = np.tile(np.array(pt.vertex), (A.shape[0], 1))
        d = np.sqrt(np.sum((A-P)**2, 1))
        return d

    def distances_to(self, pt):
        """ Return the distance from each vertex to a point. """
        d = [pt.distance(a) for a in self]
        return np.array(d)

    def greatcircles_to(self, pt):
        """ Return the great circle distances from each vertex to a point. """
        d = [pt.greatcircle(a) for a in self]
        return np.array(d)

    def nearest_point_to(self, pt):
        """ Returns the internal Point that is nearest to *pt*. If two points
        are equidistant, only one will be returned.
        """
        distances = self.distances_to(pt)
        idx = np.argmin(distances)
        return self[idx]

    def get_extents(self):
        """ Calculate a bounding box. """
        def gen_minmax(G):
            """ Get the min/max from a single pass through a generator. """
            mn = mx = next(G)
            for x in G:
                mn = min(mn, x)
                mx = max(mx, x)
            return mn, mx
        # Get the min/max for a generator defined for each dimension
        return list(map(gen_minmax,
                    map(lambda i: (c[i] for c in self.vertices),
                        range(self.rank))))

    # This code should compute the convex hull of the points and then test the
    # hull's combination space
    # Alternatively, calculate the eigenvectors, rotate, and cherrypick the
    # points
    #def max_dimension(self):
    #    """ Return the two points in the Multipoint that are furthest
    #    from each other. """
    #    dist = lambda xy0, xy1: math.sqrt((xy1[0]-xy0[0])**2 +
    #                                      (xy1[1]-xy0[1])**2)

    #    P = [(p0, p1) for p0 in self.vertices for p1 in self.vertices]
    #    D = map(dist, (p[0] for p in P), (p[1] for p in P))
    #    return P[D.index(max(D))]

    def any_within_poly(self, poly):
        """ Return whether any vertices are inside *poly* """
        for pt in self:
            if poly.contains(pt):
                return True
        return False

    def to_xyfile(self, fnm, fields=None, delimiter=' ', header=None):
        """ Write data to a delimited ASCII table.

        fnm         :   filename to write to

        kwargs:
        fields      :   specify the fields to be written (default all)

        Additional kwargs are passed to `xyfile.write_xy`.
        """
        xyfile.write_xy(self.get_vertices(), fnm, delimiter=delimiter, header=header)
        return

    def as_geojson(self, **kwargs):
        """ Print representation of internal data as a GeoJSON string.

        Parameters
        ----------
        crs : coordinate reference system
        crs_fmt : format of `crs`; may be one of ('epsg','ogc_crs_urn')
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        writer = geojson.GeoJSONWriter(self, crs=self._crs,
                    crs_fmt="ogc_crs_urn", **kwargs)
        return writer.print_json()

    def to_geojson(self, f, **kwargs):
        """ Write data as a GeoJSON string to a file-like object `f`.

        Parameters
        ----------
        f : file-like object to recieve the GeoJSON string

        *kwargs* include:
        crs : coordinate reference system
        crs_fmt : format of `crs`; may be one of ('epsg','ogc_crs_urn')
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        try:
            if not hasattr(f, "write"):
                fobj = open(f, "w")
            else:
                fobj = f
            writer = geojson.GeoJSONWriter(self, crs=self._crs,
                        crs_fmt="ogc_crs_urn", **kwargs)
            writer.write_json(fobj)
        finally:
            if not hasattr(f, "write"):
                fobj.close()
        return writer

    def to_vtk(self, f, **kwargs):
        """ Write data to an ASCII VTK .vtp file. """
        if not hasattr(f, "write"):
            with open(f, "w") as fobj:
                vtk.mp2vtp(self, fobj, **kwargs)
        else:
            vtk.mp2vtp(self, f, **kwargs)
        return

    def to_shapefile(self, fstem):
        """ Save line to a shapefile """
        if self.rank == 2:
            _shpfuncs.write_multipoint2(self, fstem)
        elif self.rank == 3:
            _shpfuncs.write_multipoint3(self, fstem)
        else:
            raise IOError("Rank must be 2 or 3 to write as a shapefile")
        return


class Multipoint(MultipointBase):
    """ Point cloud with associated attributes. This is a base class for the
    polyline and polygon classes.

    *coords*        List of 2-tuples or 3-tuples
    *data*          List, dictionary, or Metadata object for point-specific
                    data [default None]
    *properties*    Dictionary of geometry specific data [default None]
    *crs*           Coordinate reference system instance [default CARTESIAN]
    """
    _geotype = "Multipoint"

    def __init__(self, vertices, data=None, properties=None, copy_metadata=True,
                 **kwargs):
        def ispoint(a):
            return hasattr(a, "_geotype") and a._geotype == "Point"
        if len(vertices) != 0 and all(ispoint(a) for a in vertices):
            points = vertices
            crs = points[0]._crs
            rank = points[0].rank
            if len(points) != 1 and not all(pt._crs == crs for pt in points[1:]):
                raise CRSError("All points must share the same CRS")
            elif not all(rank == pt.rank for pt in points[1:]):
                raise GInitError("Input must have consistent rank")

            keys = list(points[0].data.keys())
            if len(points) != 1:
                for pt in points[1:]:
                    for key in keys:
                        if key not in pt.data:
                            keys.pop(keys.index(key))

            ptdata = {}
            for key in keys:
                ptdata[key] = [pt.data[key] for pt in points]

            if data is not None:
                ptdata.update(data)

            self.data = Metadata(ptdata, copydata=copy_metadata)
            self.vertices = [pt.vertex for pt in points]
            self.properties = properties
            self.rank = rank
            self._crs = crs

        else:

            super(Multipoint, self).__init__(vertices,
                                             data=data,
                                             properties=properties,
                                             **kwargs)
        return

    @property
    def __geo_interface__(self):
        return {"type" : "MultiPoint", "bbox" : self.bbox, "coordinates" : self.vertices}

    def within_radius(self, pt, radius):
        """ Return Multipoint of subset that is within *radius* of *pt*.
        """
        distances = self.distances_to(pt)
        indices = [i for i,d in enumerate(distances) if d <= radius]
        return self._subset(indices)

    def within_bbox(self, bbox):
        """ Return Multipoint subset that is within a square bounding box
        given by (xmin, xmax, ymin, ymax). """
        filtbbox = lambda pt: (bbox[0] <= pt.vertex[0] <= bbox[1]) and \
                              (bbox[2] <= pt.vertex[1] <= bbox[3])
        indices = [i for (i, pt) in enumerate(self) if filtbbox(pt)]
        return self._subset(indices)


class ConnectedMultipoint(MultipointBase):
    """ Class for Multipoints in which vertices are assumed to be connected. """

    def length(self):
        """ Returns the length of the line/boundary. """
        points = [Point(i, crs=self._crs) for i in self.vertices]
        distances = [a.distance(b) for a, b in zip(points[:-1], points[1:])]
        return sum(distances)

    def segments(self):
        """ Returns an generator of adjacent line segments. """
        return (self._subset((i,i+1)) for i in range(len(self)-1))

    def segment_tuples(self):
        """ Returns an generator of adjacent line segments as coordinate tuples. """
        return ((self.vertices[i], self.vertices[i+1])
                for i in range(len(self.vertices)-1))

    def intersects(self, other):
        """ Return whether an intersection exists with another geometry. """
        interxbool = (np.nan in _vecgeo.intersection(a[0][0], a[1][0], b[0][0], b[1][0],
                                                     a[0][1], a[1][1], b[0][1], b[1][1])
                    for a in self.segments() for b in other.segments())
        if self._bbox_overlap(other) and (True not in interxbool):
            return True
        else:
            return False

    def intersections(self, other, keep_duplicates=False):
        """ Return the intersections with another geometry as a Multipoint. """
        interx = (_vecgeo.intersection(a[0][0], a[1][0], b[0][0], b[1][0],
                                          a[0][1], a[1][1], b[0][1], b[1][1])
                     for a in self.segments() for b in other.segments())
        if not keep_duplicates:
            interx = set(interx)
        interx_points = []
        for vertex in interx:
            if np.nan not in vertex:
                interx_points.append(Point(vertex, properties=self.properties,
                                           crs=self._crs))
        nandata = [np.nan for _ in interx_points]
        keys = [k for k in set(list(self.data.keys()) + list(other.data.keys()))]
        d = Metadata(dict((key,nandata) for key in keys))
        return Multipoint(interx_points, data=d)

    def shortest_distance_to(self, pt):
        """ Return the shortest distance from any position on the Multipoint
        boundary to *pt* (Point).
        """
        if self._crs != pt._crs:
            raise CRSError("Point must have matching CRS")
        elif self._crs == kcrs.CARTESIAN:
            func = _vecgeo.pt_nearest_planar
        else:
            func = lambda *args: _vecgeo.pt_nearest_proj(geod, *args, tol=0.01)

        segments = list(self.segments())
        point_dist = map(func,
                         (pt.vertex for _ in segments),
                         (seg[0].vertex for seg in segments),
                         (seg[1].vertex for seg in segments))
        distances = [i[1] for i in point_dist]
        return min(distances)

    def nearest_on_boundary(self, pt):
        """ Returns the position on the Multipoint boundary that is nearest to
        pt (Point). If two points are equidistant, only one will be returned.
        """
        if self._crs != pt._crs:
            raise CRSError("Point must have matching CRS")
        elif self._crs == kcrs.CARTESIAN:
            func = _vecgeo.pt_nearest_planar
        else:
            func = lambda *args: _vecgeo.pt_nearest_proj(geod, *args, tol=0.01)

        segments = list(self.segments())
        point_dist = list(map(func,
                              (pt.vertex for _ in segments),
                              (seg[0].vertex for seg in segments),
                              (seg[1].vertex for seg in segments)))
        distances = [i[1] for i in point_dist]
        return Point(point_dist[distances.index(min(distances))][0],
                     properties=self.properties, crs=self._crs)

    def within_distance(self, pt, distance):
        """ Test whether a point is within *distance* of a ConnectedMultipoint. """
        return all(distance >= seg.shortest_distance_to(pt) for seg in self.segments())


class Line(ConnectedMultipoint):
    """ Line composed of connected vertices.

    *coords*        List of 2-tuples or 3-tuples
    *data*          List, dictionary, or Metadata object for point-specific
                    data [default None]
    *properties*    Dictionary of geometry specific data [default None]
    *crs*           Coordinate reference system instance [default CARTESIAN]
    """
    _geotype = "Line"

    @property
    def __geo_interface__(self):
        return {"type" : "LineString", "bbox" : self.bbox, "coordinates" : self.vertices}

    def add_vertex(self, vertex):
        """ Add a vertex to self.vertices. """
        if isinstance(vertex, Point):
            if self.rank == 2:
                self.vertices.append((vertex.x, vertex.y))
            elif self.rank == 3:
                self.vertices.append((vertex.x, vertex.y, vertex.z))
            self.data += vertex.data
        else:
            if self.rank == 2:
                self.vertices.append((vertex[0], vertex[1]))
            elif self.rank == 3:
                self.vertices.append((vertex[0], vertex[1], vertex[2]))
        return self

    def remove_vertex(self, index):
        """ Removes a vertex from the register by index. """
        pt = Point(self.vertices.pop(index), data=self.data[index])
        del self.data[index]
        return pt

    def extend(self, other):
        """ Combine two lines, provided that that the data formats are similar.
        """
        if self.rank == other.rank:
            if self._geotype == other._geotype:
                self.vertices.extend(other.vertices)
                self.data = self.data + other.data
            else:
                GGeoError('Cannot add inconsistent geometry types')
        else:
            GGeoError('Cannot add geometries with inconsistent rank')
        return self

    def cumlength(self):
        """ Returns the cumulative length by segment, prefixed by zero. """
        d = [0.0]
        pta = self[0]
        for ptb in self[1:]:
            d_ = pta.distance(ptb)
            d.append(d_ + d[-1])
            pta = ptb
        return d

    def subsection(self, n):
        """ Return *n* equally spaced Point instances along line. """
        segments = self.segments()
        Ltotal = self.cumlength()[-1]
        step = Ltotal / float(n-1)
        step_remaining = step

        points = [self[0]]
        x = 0.0
        pos = self[0]
        seg = next(segments)
        seg_remaining = seg.displacement()

        while x < Ltotal-1e-8:
            direction = seg[0].azimuth(seg[1])

            if step_remaining <= seg_remaining:
                pos = pos.walk(step_remaining, direction)
                x += step_remaining
                seg_remaining -= step_remaining
                step_remaining = step
                points.append(pos)
                seg.vertices[0] = pos.vertex

            else:
                pos = seg[1]
                x += seg_remaining
                step_remaining -= seg_remaining

                seg = next(segments, seg)
                seg_remaining = seg.displacement()
                # except StopIteration as e:
                #     if abs(Ltotal-x) > 1e-8:       # tolerance for endpoint
                #         raise e

        if len(points) == n-1:
            points.append(seg[-1])
        return points

    def displacement(self):
        """ Returns the distance between the first and last vertex. """
        return self[0].distance(self[-1])

    #def direction(self):
    #    """ Returns a vector of the azimuth along a line at each point,
    #    weighted by neighbour position. """
    #    # The weights are calculated to give greater weight to azimuths to
    #    # nearby points
    #    # a' = w1 a1 + w2 a2
    #    #    = (a1 h2) / (h1 + h2) + (a2 h1) / (h1 + h2)
    #    #    = (a1 h2 + a2 h1) / (h1 + h2)
    #    azs = np.empty(len(self)-1)
    #    hs = np.empty(len(self)-1)
    #    pts = [pt for pt in self]
    #    for i in range(len(self)-1):
    #        azs[i] = pt[i].azimuth(pt[i+1])
    #        hs[i] = pt[i].distance(pt[i+1])
    #    weighted_azs = np.empty(len(self))
    #    weighted_azs[0] = azs[0]
    #    weighted_azs[-1] = azs[-1]
    #    for i in range(1, len(self)-1):
    #        raise NotImplementedError("need to phase unwrap")
    #        weighted_azs[i] = (azs[i]*hs[i+1] + azs[i+1]*hs[i]) / (hs[i] + hs[i+1])
    #    return weighted_azs

    def to_polygon(self):
        """ Returns a polygon. """
        return Polygon(self.vertices)

    def to_shapefile(self, fstem):
        """ Save line to a shapefile """
        if self.rank == 2:
            _shpfuncs.write_line2(self, fstem)
        elif self.rank == 3:
            _shpfuncs.write_line3(self, fstem)
        else:
            raise IOError("rank must be 2 or 3 to write as a shapefile")
        return


class Polygon(ConnectedMultipoint):
    """ Polygon, composed of a closed sequence of vertices.

    *coords*        List of 2-tuples or 3-tuples
    *data*          List, dictionary, or Metadata object for point-specific
                    data [default None]
    *properties*    Dictionary of geometry specific data [default None]
    *subs*          List of sub-polygons [default None]
    *crs*           Coordinate reference system instance [default CARTESIAN]
    """
    _geotype = "Polygon"
    subs = []

    def __init__(self, vertices, data=None, properties=None, subs=None, **kwargs):
        ConnectedMultipoint.__init__(self, vertices, data=data, properties=properties, **kwargs)
        if vertices[0] != vertices[-1]:
            self.vertices.append(vertices[0])
        self.subs = subs if subs is not None else []
        return

    def __getitem__(self, key):
        if isinstance(key, slice):
            ind = key.indices(len(self))
            if len(self) != ((ind[1] - ind[0]) // ind[2]):
                return Line(self.vertices[key], data=self.data[key],
                            properties=self.properties, crs=self._crs)
        return super(Polygon, self).__getitem__(key)

    @property
    def __geo_interface__(self):
        return {"type" : "Polygon", "bbox" : self.bbox, "coordinates" : self.vertices}

    def _subset(self, idxs):
        """ Return a subset defined by index in *idxs*. """
        vertices = [self.vertices[i] for i in idxs]
        data = self.data.sub(idxs)
        subset = Line(vertices, data=data, properties=self.properties,
                      crs=self._crs, copy_metadata=False)
        return subset

    def perimeter(self):
        """ Return the perimeter of the polygon. If there are sub-polygons,
        their perimeters are added recursively. """
        return self.length() + sum([p.perimeter() for p in self.subs])

    def area(self):
        """ Return the two-dimensional area of the polygon. If there are
        sub-polygons, their areas are subtracted. """
        a = 0.0
        for i in range(len(self.vertices)-1):
            a += 0.5 * abs((self.vertices[i][0] + self.vertices[i+1][0])
                         * (self.vertices[i][1] - self.vertices[i+1][1]))
        return a - sum(map(lambda p: p.area(), self.subs))

    def contains(self, pt):
        """ Returns True if pt is inside or on the boundary of the
        polygon, and False otherwise. Uses a crossing number scheme.
        """
        nintx = 0
        x, y = pt[0], pt[1]
        for seg in self.segments():
            (a, b) = seg
            if _vecgeo.intersects_cn(x, y, a[0], b[0], a[1], b[1]):
                nintx += 1

        # If odd, point is inside so check subpolys
        return nintx % 2 == 1 and not any(p.contains(pt) for p in self.subs)

    def to_polyline(self):
        """ Returns a self-closing polyline. Discards sub-polygons. """
        return Line(self.vertices)

    def to_shapefile(self, fstem):
        """ Save line to a shapefile """
        if self.rank == 2:
            _shpfuncs.write_poly2(self, fstem)
        elif self.rank == 3:
            _shpfuncs.write_poly3(self, fstem)
        else:
            raise IOError("rank must be 2 or 3 to write as a shapefile")
        return


class GuppyError(Exception):
    """ Base class for guppy module errors. """
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message


class GInitError(GuppyError):
    """ Exception to raise when a guppy object fails to initialize. """
    def __init__(self, message=''):
        self.message = message


class GUnitError(GuppyError):
    """ Exception to raise there is a projected unit problem. """
    def __init__(self, message=''):
        self.message = message


class GGeoError(GuppyError):
    """ Exception to raise when a guppy object attempts an invalid transform. """
    def __init__(self, message=''):
        self.message = message


class CRSError(GuppyError):
    """ Exception to raise for invalid geodetic operations. """
    def __init__(self, message=''):
        self.message = message


def ray_intersection(pt, endpt1, endpt2, direction=0.0):
    """ Determines whether a ray intersects a line segment. If yes,
    returns the point of intersection. If no, return None. Input
    "points" should be tuples or similar. Input direction is in
    radians, and defines the ray path from pt.
    """
    m_ray = math.tan(direction)
    if endpt2[0] != endpt1[0]:
        m_lin = float(endpt2[1] - endpt1[1]) / float(endpt2[0] - endpt1[0])
        if m_ray == m_lin:      # Lines are parallel
            return
        else:
            x_int = ( (m_ray * pt[0] - m_lin * endpt1[0] - pt[1] + endpt1[1])
                / (m_ray - m_lin) )

        # Test that y_int is within segment and ray points toward segment
        if ( (x_int >= endpt1[0]) is not (x_int >= endpt2[0]) and
            (x_int-pt[0] > 0) is (math.cos(direction) > 0) ):
            y_int = ( (m_ray * m_lin * pt[0] - m_ray * m_lin * endpt1[0]
                + m_ray * endpt1[1] - m_lin * pt[1]) / (m_ray - m_lin) )
            return (x_int, y_int)
        else:
            return

    else:       # Line segment is vertical
        if direction % math.pi/2. == 0.0 and direction != 0.0:
            # Lines are parallel
            return
        x_int = float(endpt1[0])
        y_int = (x_int - pt[0]) * m_ray + pt[1]
        # Test that y_int is within segment and ray points toward segment
        if ( (y_int >= endpt1[1]) is not (y_int >= endpt2[1]) and
            (x_int-pt[0] > 0) is (math.cos(direction) > 0) ):
            return (x_int, y_int)
        else:
            return


def greatcircle(pta, ptb, method="vicenty"):
    """ Computes the great circle distance between n point pairs on a
    sphere. Returns a list of length (n-1)

    [pntlist] contains a list of point objects

    [method] may be "vicenty" (default) or "haversine". The Haversine
    method is roughly 20% faster, but may yield rounding errors when
    coordinates are antipodal.
    """

    radius = 6371.
    deg2rad = np.pi / 180.

    x1 = pta.x * deg2rad
    x2 = ptb.x * deg2rad
    y1 = pta.y * deg2rad
    y2 = ptb.y * deg2rad

    dx = x2 - x1
    dy = y2 - y1

    if method == "haversine":
        try:
            distance = 2 * radius * math.asin(math.sqrt((math.sin(dy /
                2.))**2 + math.cos(y1) * math.cos(y2) *
                (math.sin(dx / 2.))**2))
        except GGeoError:
            traceback.print_exc()

    elif method == "vicenty":
        try:
            a = math.sqrt((math.cos(y2) * math.sin(dx))**2 +
                (math.cos(y1) * math.sin(y2) - math.sin(y1) *
                math.cos(y2) * math.cos(dx))**2)
            b = (math.sin(y1) * math.sin(y2) + math.cos(y1) *
                math.cos(y2) * math.cos(dx))
            distance = radius * math.atan2(a, b)
        except ZeroDivisionError:
            raise GGeoError("Zero in denominator")
            return None
        except:
            traceback.print_exc()
    else:
        raise Exception("Distance method unrecognized")
        distance = np.nan

    return distance


def sortby(A, B):
    """ Sort a list A by the values in an ordered list B. """
    if len(A) != len(B):
        raise GGeoError("A and B must be of the same length")
    comb = zip(B,A)
    comb.sort()
    return [i[1] for i in comb]


def points_to_multipoint(points):
    """ Merge *points* into a Multipoint instance. Point properties are stored
    as Multipoint data. All points must use the same CRS.
    """
    crs = points[0]._crs
    if not all(pt._crs == crs for pt in points):
        raise CRSError("All points must share the same CRS")

    keys = list(points[0].properties.keys())
    for pt in points[1:]:
        for key in keys:
            if key not in pt.properties:
                keys.pop(keys.index(key))

    ptdata = {}
    for key in keys:
        ptdata[key] = [pt.properties[key] for pt in points]

    vertices = [pt.vertex for pt in points]

    return Multipoint(vertices, data=ptdata, crs=crs)




""" Convenience reader functions """

import os
from numbers import Number
import shapefile
from . import guppy
from . import geojson
from . import xyfile
from .. import crs as kcrs
from .metadata import Metadata

def _parsegeojsoncrs(jsoncrs, default=kcrs.LONLAT_WGS84):
    """ From a tuple representing a GeoJSON (name,None) or (href,type) pair,
    return an appropriate karta crs instance. """
    if jsoncrs is None:
        return default
    elif jsoncrs["type"] == "name":
        for c in kcrs.crslist:
            if c.urn == jsoncrs["properties"]["name"]:
                return c
        return kcrs.CRS("unknown", "unknown", jsoncrs["properties"]["name"])
    elif jsoncrs["type"] == "link":
        return kcrs.CRS("unknown", jsoncrs["properties"]["type"], 
                                   jsoncrs["properties"]["href"])
    else:
        raise TypeError("Invalid GeoJSON CRS type")

def read_geojson(f):
    """ Read a GeoJSON object file and return a list of geometries """

    def convert_geometry(geom, defaultcrs=None, **kw):
        coordsys = _parsegeojsoncrs(geom.crs, default=defaultcrs)
        if isinstance(geom, geojson.Point):
            return guppy.Point(geom.coordinates, crs=coordsys, **kw)
        elif isinstance(geom, geojson.MultiPoint):
            return guppy.Multipoint(geom.coordinates, crs=coordsys, **kw)
        elif isinstance(geom, geojson.LineString):
            return guppy.Line(geom.coordinates, crs=coordsys, **kw)
        elif isinstance(geom, geojson.Polygon):
            return guppy.Polygon(geom.coordinates[0],
                                 subs=geom.coordinates[1:],
                                 crs=coordsys, **kw)
        else:
            raise TypeError("Argument to convert_geometry is a not a JSON geometry")

    def convert_feature(feat, **kw):
        data = feat.properties["vector"]
        for key, val in data.items():
            if any(isinstance(a, Number) or hasattr(a, "dtype") for a in val):
                for i in range(len(val)):
                    if val[i] is None:
                        val[i] = float('nan')
        prop = feat.properties["scalar"]
        return convert_geometry(feat.geometry, data=data, properties=prop, **kw)

    R = geojson.GeoJSONReader(f)
    gplist = []
    for item in R.items():
        if isinstance(item, geojson.Feature):
            gplist.append(convert_feature(item))
        elif isinstance(item, geojson.FeatureCollection):
            for feature in item.features:
                gplist.append(convert_feature(feature, defaultcrs=item.crs))
        else:
            gplist.append(convert_geometry(item))
    return gplist

def _geojson_properties2guppy(properties, n):
    """ Takes a dictionary (derived from a GeoJSON properties object) and
    divides it into singleton properties and *n*-degree data. """
    props = {}
    data = {}
    for (key, value) in properties.items():
        if isinstance(value, list) or isinstance(value, tuple):
            if len(value) == n:
                data[key] = value
            else:
                raise ValueError("properties must be singleton or per-vertex")
        else:
            props[key] = value
    return props, data

def read_xyfile(f, delimiter='', header_rows=0, astype=guppy.Multipoint, coordrank=2):
    """ Read an ASCII delimited table and return a guppy object given by *astype*.
    """
    dat = xyfile.load_xy(f, delimiter=delimiter, header_rows=header_rows)
    ncols = dat.shape[1]
    if ncols >= coordrank:
        coords = dat[:,:coordrank]
        if ncols > coordrank:
            data = dat[:,coordrank:]
        else:
            data = None
        return astype(coords, data=data)
    else:
        raise IOError('data table has insufficient number of columns')

### Shapefile functions ###

def get_filenames(stem, check=False):
    """ Given a filename basename, return the associated shapefile paths. If
    `check` is True, ensure that the files exist."""
    shp = stem + '.shp'
    shx = stem + '.shx'
    dbf = stem + '.dbf'
    if check:
        for fnm in (shp, shx, dbf):
            if not os.path.isfile(fnm):
                raise Exception('missing {0}'.format(fnm))
    return {'shp':shp, 'shx':shx, 'dbf':dbf}

def open_file_dict(fdict):
    """ Open each file in a dictionary of filenames and return a matching
    dictionary of the file objects. """
    files = {}
    for ext in fdict.keys():
        files[ext] = open(fdict[ext], 'rb')
    return files

dBase_type_dict = {"I": int,
                   "O": float,
                   "C": str,
                   "@": lambda a: a,    # Temporary
                   "D": lambda a: a,
                   "L": bool}

def recordsasdata(reader):
    """ Interpret shapefile records as a Metadata object """
    d = {}
    idfunc = lambda a: a
    records = [rec for rec in reader.records()]
    for (i,k) in enumerate(reader.fields[1:]):
        f = dBase_type_dict.get(k[1], idfunc)
        d[k[0]] = [f(rec[i]) for rec in records]
    return Metadata(d)

def recordsasproperties(reader):
    """ Interpret shapefile records as a list of properties dictionaries """
    proplist = []
    keys = reader.fields
    idfunc = lambda a: a
    for (i,rec) in enumerate(reader.records()):
        properties = {}
        for (k,v) in zip(keys, rec):
            f = dBase_type_dict.get(k[1], idfunc)
            properties[k[0]] = f(v)
        proplist.append(properties)
    return proplist

def read_shapefile(stem, crs=None):
    """ Read a shapefile given `stem`, which is the name without an extension.
    The geometry CRS must be specified, otherwise it is assumed to be
    cartesian.
    """
    if crs is None:
        crs = kcrs.CARTESIAN
    fnms = get_filenames(stem, check=True)

    try:
        files = open_file_dict(fnms)
        reader = shapefile.Reader(shp=files['shp'], shx=files['shx'],
                                  dbf=files['dbf'])

        if reader.shapeType == 1:       # Points
            verts = [shp.points[0] for shp in reader.shapes()]
            d = recordsasdata(reader)
            geoms = [guppy.Multipoint(verts, data=d, crs=crs)]

        elif reader.shapeType == 3:     # Lines
            plist = recordsasproperties(reader)
            geoms = []
            for (shp,prop) in zip(reader.shapes(), plist):
                parts = list(shp.parts)
                parts.append(len(shp.points))
                for (i0, i1) in zip(parts[:-1], parts[1:]):
                    points = shp.points[i0:i1]
                    geoms.append(guppy.Line(points, properties=prop, crs=crs))

        elif reader.shapeType == 5:     # Polygon
            plist = recordsasproperties(reader)
            geoms = []
            for (shp,prop) in zip(reader.shapes(), plist):
                parts = list(shp.parts)
                parts.append(len(shp.points))
                for (i0, i1) in zip(parts[:-1], parts[1:]):
                    points = shp.points[i0:i1]
                    geoms.append(guppy.Polygon(points, properties=prop, crs=crs))

        else:
            raise NotImplementedError("Shapefile shape type {0} not "
                "implemented".format(reader.shapeType))

    finally:
        for f in files.values():
            f.close()

    return geoms




# QuadTree functions

def iswithin(tuple bbox, tuple pt):
    """ Return whether a point is within a bounding box (planar approximation). """
    cdef float xmn = bbox[0]
    cdef float xmx = bbox[1]
    cdef float ymn = bbox[2]
    cdef float ymx = bbox[3]
    cdef float x = pt[0]
    cdef float y = pt[1]
    if (xmn <= x < xmx and ymn <= y < ymx):
        return True
    else:
        return False

def qthashpt(float xmin, float xmax, float ymin, float ymax, float x, float y):
    """ Returns a generator that returns successive quadrants [0-3] that
    constitute a geohash for *pt* in a global *bbox*. """
    cdef float xm, ym
    cdef int geohash
    while True:
        xm = 0.5 * (xmin + xmax)
        ym = 0.5 * (ymin + ymax)
        if x < xm:
            if y < ym:
                geohash = 0
                xmax = xm
                ymax = ym
            elif y >= ym:
                geohash = 2
                xmax = xm
                ymin = ym
            else:
                raise HashError
        elif x >= xm:
            if y < ym:
                geohash = 1
                xmin = xm
                ymax = ym
            elif y >= ym:
                geohash = 3
                xmin = xm
                ymin = ym
            else:
                raise HashError
        else:
            raise HashError
        yield geohash
        
class HashError(Exception):
    pass




import os
from distutils.core import setup
from distutils.extension import Extension
import numpy

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
    include_dirs = [numpy.get_include()]
    ext = ".pyx"
except ImportError:
    USE_CYTHON = False
    include_dirs = []
    ext = ".c"
    print("Warning: Cython not imported")
    print("If C sources exist in the source tree, they will be used")

extensions = [Extension("karta.raster.cfill_sinks",
                        ["karta/raster/cfill_sinks"+ext],
                        include_dirs=include_dirs),
              Extension("karta.raster.crfuncs",
                        ["karta/raster/crfuncs"+ext],
                        include_dirs=include_dirs),
              Extension("karta.vector._cvectorgeo",
                        ["karta/vector/_cvectorgeo"+ext],
                        include_dirs=include_dirs)]

if USE_CYTHON:
    extensions = cythonize(extensions)

for extension in extensions:
    if not all(os.path.exists(src) for src in extension.sources):
        # C-extension sources missing, so don't try to build them
        extensions = []
        print("Warning: Extension source not found: {0}".format(extension.sources))
        print("Not building accelerated modules")
        break

setup(
    name = "karta",
    version = "0.3.0",
    author = "Nat Wilson",
    author_email = "njwilson23@gmail.com",
    packages = ["karta", "karta.vector", "karta.raster"],
    url = "http://njwilson23.github.io/karta/",
    description = "A tidy Python package for geospatial computation",
    long_description = """
Karta - tidy Python package for geospatial computation
======================================================

*Karta* is a Leatherman for geographic analyses. *Karta* provides an
lightweight interface for solving problems in Python/Python3. It presents a
simple and clean vector and raster data types, a small selection of
geographical analysis functions, and the ability to read and write several
useful formats.

Goals of *Karta* include exposing a simple and fast framework for spatial
analysis. *Karta* is under development and suggestions and pull requests are
welcome, particularly to improve format support and test coverage.

DOCUMENTATION
-------------

API documentation can be built by running ``make`` from the ``doc/`` directory.
The beginnings of a tutorial to introduce *Karta* are in the `Wiki
<https://github.com/njwilson23/karta/wiki/Tutorial>`__.

CONTENTS AT A GLANCE
--------------------

-  vector

   -  guppy: Vector geometry classes (e.g. ``Point``, ``Multipoint``,
      ``Line``, ``Polygon``)
   -  gpx: GPX class for parsing and constructing GPX eXchange files
   -  geojson: Classes and functions for reading and writing GeoJSON
   -  vtk: XML-based VTK interface
   -  shp\_funcs: Shapefile-to-guppy conversions through *pyshp*
      interface
   -  stats: Basic geostatistical functions
   -  xyfile: ASCII table functions
   -  quadtree: Simple QuadTree implementation

-  raster

   -  grid: Basic Grid types, including ``StructuredGrid`` and
      ``RegularGrid``
   -  raster: General purpose raster functions
   -  aaigrid: Grid subclass specifically for reading and writing ESRI
      ASCII grids
   -  flow: Stream flow functions
   -  streamline: Streamline calculation

-  tests : unit tests

FORMATS
-------

*Karta* provides a basic working interface to several of common file
formats. Currently partially-supported are:

-  vector

   -  GeoJSON (r,w)
   -  ESRI Shapefiles via pyshp (r,w)
   -  GPS eXchange (GPX) (r,w)
   -  ASCII tables (XYZ) (r,w)
   -  VTK (w)

-  raster

   -  ESRI ASCII Grid (r,w)
   -  GeoTiff through GDAL (r)
   -  USGS DEM (WIP)

INSTALLATION
------------

The easiest way to install is to use ``pip``.

To install the latest release from PyPI, run

::

    pip install karta

To build from source,

::

    git clone https://github.com/njwilson23/karta.git
    pip install -r karta/requirements.txt
    pip install karta/

DEPENDENCIES
------------

Required
~~~~~~~~

-  Python 2.7+ or Python 3.2+
-  numpy
-  pyshp
-  pyproj (required for geodetic calculations)

Optional
~~~~~~~~

-  scipy
-  gdal
-  cython

Cython is an optional dependency used to speed up select functions. In general,
enhanced-performance functions will then be called automatically when
available, otherwise *Karta* will fall back to numpy and pure-Python versions.

When installing from PyPI, C source code is provided and will be automatically
compiled if a suitable compiler is available.

TESTING
-------

To run all unit tests, execute

::

    python tests/runtests.py

LICENSE
-------

This software is provided under the MIT license.

MIT License:
~~~~~~~~~~~~

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

""",
    download_url = "https://github.com/njwilson23/karta/archive/master.zip",
    classifiers = ["Programming Language :: Python",
                   "Programming Language :: Python :: 3",
                   "Development Status :: 3 - Alpha",
                   "Topic :: Scientific/Engineering"],
    license = "MIT License",
    ext_modules = extensions
)




""" This is a database of CRS IDs. """

from collections import namedtuple

# "name" is a human readable identifier, and should not be used as a hash
# "type" is one of ("local", "geographical", "projected")
# "urn" is an Open Geospatial Consortium uniform resource name
CRS = namedtuple("CRS", ["name", "type", "urn"])


# Local and geographical

CARTESIAN = CRS("cartesian", "local", "urn:ogc:def:crs:EPSG::5806")
LONLAT_WGS84 = CRS("lonlat (WGS 84)", "geographical", "urn:ogc:def:crs:EPSG::4326")
LONLAT_NAD27 = CRS("lonlat (NAD 27)", "geographical", "urn:ogc:def:crs:EPSG::4267")
LONLAT_NAD83 = CRS("lonlat (NAD 83)", "geographical", "urn:ogc:def:crs:EPSG::4269")


# UTM zones
UTM1N = CRS("UTM Zone 1 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32601")
UTM1S = CRS("UTM Zone 1 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32701")
UTM2N = CRS("UTM Zone 2 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32602")
UTM2S = CRS("UTM Zone 2 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32702")
UTM3N = CRS("UTM Zone 3 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32603")
UTM3S = CRS("UTM Zone 3 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32703")
UTM4N = CRS("UTM Zone 4 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32604")
UTM4S = CRS("UTM Zone 4 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32704")
UTM5N = CRS("UTM Zone 5 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32605")
UTM5S = CRS("UTM Zone 5 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32705")
UTM6N = CRS("UTM Zone 6 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32606")
UTM6S = CRS("UTM Zone 6 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32706")
UTM7N = CRS("UTM Zone 7 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32607")
UTM7S = CRS("UTM Zone 7 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32707")
UTM8N = CRS("UTM Zone 8 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32608")
UTM8S = CRS("UTM Zone 8 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32708")
UTM9N = CRS("UTM Zone 9 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32609")
UTM9S = CRS("UTM Zone 9 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32709")
UTM10N = CRS("UTM Zone 10 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32610")
UTM10S = CRS("UTM Zone 10 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32710")
UTM11N = CRS("UTM Zone 11 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32611")
UTM11S = CRS("UTM Zone 11 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32711")
UTM12N = CRS("UTM Zone 12 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32612")
UTM12S = CRS("UTM Zone 12 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32712")
UTM13N = CRS("UTM Zone 13 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32613")
UTM13S = CRS("UTM Zone 13 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32713")
UTM14N = CRS("UTM Zone 14 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32614")
UTM14S = CRS("UTM Zone 14 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32714")
UTM15N = CRS("UTM Zone 15 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32615")
UTM15S = CRS("UTM Zone 15 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32715")
UTM16N = CRS("UTM Zone 16 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32616")
UTM16S = CRS("UTM Zone 16 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32716")
UTM17N = CRS("UTM Zone 17 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32617")
UTM17S = CRS("UTM Zone 17 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32717")
UTM18N = CRS("UTM Zone 18 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32618")
UTM18S = CRS("UTM Zone 18 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32718")
UTM19N = CRS("UTM Zone 19 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32619")
UTM19S = CRS("UTM Zone 19 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32719")
UTM20N = CRS("UTM Zone 20 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32620")
UTM20S = CRS("UTM Zone 20 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32720")
UTM21N = CRS("UTM Zone 21 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32621")
UTM21S = CRS("UTM Zone 21 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32721")
UTM22N = CRS("UTM Zone 22 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32622")
UTM22S = CRS("UTM Zone 22 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32722")
UTM23N = CRS("UTM Zone 23 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32623")
UTM23S = CRS("UTM Zone 23 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32723")
UTM24N = CRS("UTM Zone 24 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32624")
UTM24S = CRS("UTM Zone 24 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32724")
UTM25N = CRS("UTM Zone 25 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32625")
UTM25S = CRS("UTM Zone 25 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32725")
UTM26N = CRS("UTM Zone 26 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32626")
UTM26S = CRS("UTM Zone 26 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32726")
UTM27N = CRS("UTM Zone 27 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32627")
UTM27S = CRS("UTM Zone 27 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32727")
UTM28N = CRS("UTM Zone 28 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32628")
UTM28S = CRS("UTM Zone 28 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32728")
UTM29N = CRS("UTM Zone 29 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32629")
UTM29S = CRS("UTM Zone 29 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32729")
UTM30N = CRS("UTM Zone 30 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32630")
UTM30S = CRS("UTM Zone 30 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32730")
UTM31N = CRS("UTM Zone 31 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32631")
UTM31S = CRS("UTM Zone 31 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32731")
UTM32N = CRS("UTM Zone 32 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32632")
UTM32S = CRS("UTM Zone 32 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32732")
UTM33N = CRS("UTM Zone 33 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32633")
UTM33S = CRS("UTM Zone 33 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32733")
UTM34N = CRS("UTM Zone 34 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32634")
UTM34S = CRS("UTM Zone 34 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32734")
UTM35N = CRS("UTM Zone 35 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32635")
UTM35S = CRS("UTM Zone 35 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32735")
UTM36N = CRS("UTM Zone 36 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32636")
UTM36S = CRS("UTM Zone 36 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32736")
UTM37N = CRS("UTM Zone 37 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32637")
UTM37S = CRS("UTM Zone 37 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32737")
UTM38N = CRS("UTM Zone 38 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32638")
UTM38S = CRS("UTM Zone 38 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32738")
UTM39N = CRS("UTM Zone 39 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32639")
UTM39S = CRS("UTM Zone 39 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32739")
UTM40N = CRS("UTM Zone 40 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32640")
UTM40S = CRS("UTM Zone 40 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32740")
UTM41N = CRS("UTM Zone 41 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32641")
UTM41S = CRS("UTM Zone 41 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32741")
UTM42N = CRS("UTM Zone 42 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32642")
UTM42S = CRS("UTM Zone 42 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32742")
UTM43N = CRS("UTM Zone 43 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32643")
UTM43S = CRS("UTM Zone 43 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32743")
UTM44N = CRS("UTM Zone 44 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32644")
UTM44S = CRS("UTM Zone 44 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32744")
UTM45N = CRS("UTM Zone 45 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32645")
UTM45S = CRS("UTM Zone 45 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32745")
UTM46N = CRS("UTM Zone 46 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32646")
UTM46S = CRS("UTM Zone 46 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32746")
UTM47N = CRS("UTM Zone 47 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32647")
UTM47S = CRS("UTM Zone 47 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32747")
UTM48N = CRS("UTM Zone 48 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32648")
UTM48S = CRS("UTM Zone 48 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32748")
UTM49N = CRS("UTM Zone 49 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32649")
UTM49S = CRS("UTM Zone 49 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32749")
UTM50N = CRS("UTM Zone 50 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32650")
UTM50S = CRS("UTM Zone 50 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32750")
UTM51N = CRS("UTM Zone 51 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32651")
UTM51S = CRS("UTM Zone 51 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32751")
UTM52N = CRS("UTM Zone 52 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32652")
UTM52S = CRS("UTM Zone 52 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32752")
UTM53N = CRS("UTM Zone 53 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32653")
UTM53S = CRS("UTM Zone 53 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32753")
UTM54N = CRS("UTM Zone 54 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32654")
UTM54S = CRS("UTM Zone 54 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32754")
UTM55N = CRS("UTM Zone 55 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32655")
UTM55S = CRS("UTM Zone 55 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32755")
UTM56N = CRS("UTM Zone 56 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32656")
UTM56S = CRS("UTM Zone 56 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32756")
UTM57N = CRS("UTM Zone 57 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32657")
UTM57S = CRS("UTM Zone 57 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32757")
UTM58N = CRS("UTM Zone 58 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32658")
UTM58S = CRS("UTM Zone 58 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32758")
UTM59N = CRS("UTM Zone 59 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32659")
UTM59S = CRS("UTM Zone 59 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32759")
UTM60N = CRS("UTM Zone 60 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32660")
UTM60S = CRS("UTM Zone 60 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32760")


# Polar stereographic

UPSNORTH_WGS84 = CRS("UPS North", "projected", "urn:ogc:def:crs:EPSG::32661")
UPSSOUTH_WGS84 = CRS("UPS South", "projected", "urn:ogc:def:crs:EPSG::32761")

NSIDCNORTH = CRS("NSIDC Sea Ice Sterographic North", "projected", "urn:ogc:def:crs:EPSG::3411")
NSIDCSOUTH = CRS("NSIDC Sea Ice Sterographic South", "projected", "urn:ogc:def:crs:EPSG::3412")


# Convenience aliases
LONLAT = LONLAT_WGS84
UPSNORTH = UPSNORTH_WGS84
UPSSOUTH = UPSSOUTH_WGS84



localnames = locals()
crslist = [c for c in localnames.values() if isinstance(c, CRS)]

